/*
Subject: Katmai: IsDescendant and index seeks/scans for hierarchyid
Considering the NewOrg sample table from Katmai Books Online (see: 
ms-help://MS.SQLCC.v10/MS.SQLSVR.v10.en/s10de_1devconc/html/fd943d84-dbe6-4a05-912b-c88164998d80.htm), 
I expect the following query to perform an index seek:
*/

SELECT COUNT(*)
FROM HumanResources.NewOrg b
WHERE CAST('/1/1/4/' AS hierarchyid).IsDescendant(b.OrgNode)=1

-- however, the above query performs an index scan (using 9 logical reads), 
-- although the following query:

SELECT COUNT(*)
FROM HumanResources.NewOrg b, HumanResources.NewOrg a
WHERE a.OrgNode=CAST('/1/1/4/' AS hierarchyid)
AND a.OrgNode.IsDescendant(b.OrgNode)=1

/*
performs two index seeks (using only 5 logical reads).

I think this is a case where the query optimizer can do a better job 
(probably it's just a simple scenario that was overlooked in the July CTP 
version). Should I open a bug/suggestion on Connect ?
Razvan

Ok, I see what you're saying. I'm pleasantly surprised they put
this logic in, too (insofar as it gets used)!

I think the seek might have been considered and it ended
up costing more due to bad rowcount estimates.

On my machine (different data in the table, I think), if I try
*/
SELECT COUNT(*)
FROM HumanResources.NewOrgb, HumanResources.NewOrga
WHERE a.OrgNode=CAST('/1/1/4/' AS hierarchyid)
AND CAST('/1/1/4/' AS hierarchyid).IsDescendant(b.OrgNode)=1

/*
I get a scan on the bottom branch, not a seek. It seems
like the reason is that there's a way-too-high rowcount
estimate (95 rows out of the 290 in my table) for the
filter on hierarchyid::Parse('/1/1/4/').IsDescendant(OrgNode)=1.

For the query you posted, I get a rowcount of 26 for the
condition a.OrgNode.IsDescendant(b.OrgNode)=1. The actual
rowcount is 4.

I'm not sure how the rowcount estimates are calculated here,
but I would have hoped the better estimate would have been
with <constant expression>.IsDescendant(), not with
<other table's column>.IsDescendant. But as far as I can
tell, the rowcount estimate for
<constant expression>.IsDescendant() = 1 is a generic 1/3
of the table rows, regardless of <constant expression>.

Hopefully for Katmai, or eventually, they can improve this,
either by using smaller estimates when <constant expression>
is at a deep level, or by leveraging some of the existing
strategies that handle estimates for LIKE conditions.
*/

-- The optimizer probably should have rewritten the joined query as

SELECT COUNT(*)
FROM HumanResources.NewOrg
WHERE CAST('/1/1/4/' AS hierarchyid).IsDescendant(b.OrgNode)=1
AND EXISTS (
  SELECT * FROM HumanResources.NewOrg
  WHERE CAST('/1/1/4/' AS hierarchyid) = OrgNode
)
/*
but until the rowcount estimates are improved, that would have
made things worse. (If you look at the plan for this, note that
the upper and lower branches into the loop are reversed.)
SK

You can post this on Connect, but I don't see how the optimizer
can easily figure out how to use a seek here without having a
lot of special logic that would depend on much more than what
the CLR method IsDescendant exposes. While that might help some
standard queries out, it would be a dangerous choice, I think.

If there's no scan for the first query, that means that the
IsDescendant method is not called on every b.OrgNode value.
The only way for the optimizer to do this is to "understand"
what IsDescendant = 1 means (i.e., have special logic for
this.

Am I missing something here, and is this a simpler scenario
thank I realize?

The second query can use a seek because a.OrgNode = a constant.

That's the first seek (and it's obvious, indeed), but I was talking about 
the second seek (from the second query).

The only way for the optimizer to do this is to "understand"
what IsDescendant = 1 means (i.e., have special logic for
this.

The optimizer understands what IsDescendant=1 means: look closely at the 
second index seek for the second query. It says:

Object: [AdventureWorks].[HumanResources].[NewOrg].[PK_NewOrg_OrgNode] 
[b]
Seek Predicates: Start Range: 
[AdventureWorks].[HumanResources].[NewOrg].OrgNode >= Scalar 
Operator([AdventureWorks].[HumanResources].[NewOrg].[OrgNode] as 
[a].[OrgNode]), End Range: 
[AdventureWorks].[HumanResources].[NewOrg].OrgNode <= Scalar 
Operator([Expr1005])

In fact, the second query is simply the same as the first query: instead 
of checking "b is child of x" (where x is a constant), I am checking "b 
is child of a and a=x". I see no obvious reason that the second query 
performs better then the first (in fact, the first should perform 
better).
*/